from kgcnn.data.transform.scaler.mol import *
import logging

# For old reference.
logging.error(
    "Module '%s' is deprecated and will be removed in future versions. Please move to 'kgcnn.data.transform.scaler'." % __name__)